# -*- coding:utf-8 -*-
# Author: Bingning Wang (王炳宁)
# Email: research@bingning.wang
# HomePage: www.bingning.wang
# Date: 16-10-14 下午3:06
import lasagne

from public_functions import *
from theanoUtils import *
import theano

rng = np.random.RandomState(1991)


def add_to_parameter(obj, name, value):
    if not hasattr(obj, name):
        obj.__setattr__(name, value)


class RNN(object):
    def __init__(self,
                 n_in=None,
                 weight_initiation='svd',
                 n_hidden=50,
                 only_return_final=False,
                 return_method='last',
                 backwards=False,
                 ignore_zero=False,
                 learn_hidden_init=False,
                 batch_mode=False,
                 dtype=theano.config.floatX,
                 name='RNN',
                 inner_activation=T.tanh,
                 **kwargs
                 ):
        self.name = name
        self.return_method = return_method
        self.inner_activation = inner_activation
        self.learn_hidden_init = learn_hidden_init
        self.ignore_zero = ignore_zero
        self.backwards = backwards
        self.only_return_final = only_return_final
        self.n_hidden = n_hidden
        self.W_initiation = weight_initiation
        self.n_in = n_in
        self.dtype = dtype
        self.batch_mode = batch_mode
        # parameter
        add_to_parameter(self, 'b_h', theano.shared(np.zeros(n_hidden, dtype=self.dtype), name=self.name + '_b_h'))
        add_to_parameter(self, 'W_ih', self.init_one_parameter(shape=[n_in, n_hidden], name=self.name + '_W_ih'))
        add_to_parameter(self, 'W_hh', self.init_one_parameter(shape=[n_hidden, n_hidden], name=self.name + '_W_hh'))
        add_to_parameter(self, '_h0', theano.shared(np.zeros(n_hidden, dtype=self.dtype), name=self.name + '_h0'))
        if not hasattr(self, 'params'):
            self.params = [self.W_ih, self.W_hh, self.b_h]
            if learn_hidden_init:
                self.params.append(self._h0)
        self.h_vals = None
        self.__updates = None

    def set_h0(self, h0):
        self._h0 = h0

    def reset_h0(self):
        self._h0 = theano.shared(np.zeros(self.n_hidden, dtype='float32'), name=self.name + '_h0')

    @property
    def h0(self):
        return self._h0

    def init_one_parameter(self, shape=None, name='None'):
        if shape is None:
            shape = []
        if len(shape) == 2:
            return theano.shared(self.sample_weights(shape[0], shape[1]), name=name)
        return theano.shared(np.float32(np.random.normal(0.0, 0.1, shape[0])), name=name)

    def sample_weights(self, sizeX, sizeY):
        """
        it has been proved that the max singular value of a matirx can not
        exceed 1 for the non exploding RNN issues
        :param sizeY: the initiation matrix size y
        :param sizeX:the initiation matrix size x
        :return: the svd matrix remove max values
        """
        scale = np.sqrt(2.0 / (sizeX + sizeY))
        if self.W_initiation == 'random':
            return np.asanyarray(np.random.normal(size=[sizeX, sizeY]), dtype=self.dtype)
        else:
            # values = np.ndarray([sizeX, sizeY], dtype=self.dtype)
            # for dx in xrange(sizeX):
            #     vals = np.random.uniform(low=-0.5, high=0.5, size=(sizeY,))
            #     values[dx, :] = vals
            # _, svs, _ = np.linalg.svd(values)
            # # svs[0] is the largest singular value
            # values = values / svs[0]
            values = np.asanyarray(np.random.normal(loc=0.0, scale=scale, size=[sizeX, sizeY]), dtype='float32')
            return values

    @property
    def updates(self):
        return self.__updates

    def __call__(self, inputs, batch_mode=False, **kwargs):
        self.batch_mode = batch_mode
        if batch_mode:
            self.sample_size = inputs.shape[0]
            inputs = inputs.dimshuffle(1, 0, 2)
        outputs_info = self.get_initiation()
        non_sequence = self.get_sequences()
        step_fun = self.one_step
        self.h_vals = None
        self.h_vals, self.__updates = theano.scan(fn=step_fun,
                                                  sequences=inputs,
                                                  outputs_info=outputs_info,
                                                  go_backwards=self.backwards,
                                                  non_sequences=non_sequence)
        return self.hidden_states

    def one_step(self, x_t, h_tm1, W_ih, W_hh, b_h):
        h_t = self.inner_activation(theano.dot(x_t, W_ih) + theano.dot(h_tm1, W_hh) + b_h)
        return h_t

    @property
    def hidden_states(self):
        if self.only_return_final:
            if self.return_method == 'ave':
                return T.mean(self.h_vals, axis=0)
            elif self.return_method == 'max':
                return T.max(self.h_vals, axis=0)
            else:
                return self.h_vals[-1]
        else:
            if self.backwards:
                self.h_vals = self.h_vals[::-1]
            if self.batch_mode:
                self.h_vals = self.h_vals.dimshuffle(1, 0, 2)
        return self.h_vals

    @property
    def parameter(self):
        return self.params

    def get_initiation(self):
        if self.batch_mode:
            return [T.alloc(self.h0, self.sample_size, self.n_hidden)]
        return [self.h0]

    def get_sequences(self):
        return [self.W_ih, self.W_hh, self.b_h]


class RNN_Decoder(RNN):
    def __init__(self,
                 e_d_shape_match,
                 word_embedding=None,
                 n_encoder=None,
                 **kwargs):
        RNN.__init__(self, **kwargs)
        self.e_d_shape_match = e_d_shape_match
        self.n_encoder = n_encoder
        if e_d_shape_match is not True:
            if n_encoder is None:
                assert 'you must specify the encoder dimension'
            else:
                self.en_de_projection = self.init_one_parameter(shape=[n_encoder, self.n_hidden],
                                                                name=self.name + 'en_de_projection')
                self.params.append(self.en_de_projection)

        self.projection_matrix_hidden = self.init_one_parameter(shape=[self.n_hidden, self.n_in],
                                                                name=self.name + 'projection_matrix_hidden')
        self.params.append(self.projection_matrix_hidden)
        self.mode = 'Test'
        self.start_word = T.constant(-1, dtype='int32')
        self.start_word_embedding = theano.shared(np.zeros(self.n_in, dtype=self.dtype), name=self.name + '_x0_')
        self.word_embedding = word_embedding
        if self.word_embedding is not None:
            self.stop_word_id = word_embedding.shape[0] - 1

    def set_start_word_embedding(self, start_we):
        self.start_word_embedding = start_we

    def set_word_embedding(self, wordembedding):
        self.word_embedding = wordembedding

    def decode(self, start_hidden, max_length, true_word=None):
        if self.e_d_shape_match:
            self.set_h0(start_hidden)
        else:
            encoder_input = T.tanh(T.dot(start_hidden, self.en_de_projection))
            self.set_h0(encoder_input)
        if true_word is not None:
            self.mode = 'Train'
        else:
            self.mode = 'Test'
            true_word = T.arange(0, max_length, dtype='int32')
        outputs_info = self.get_initiation()
        non_sequence = self.get_sequences()
        step_fun = self.one_step
        self.h_vals = None
        self.h_vals, self.__updates = theano.scan(fn=step_fun,
                                                  sequences=true_word,
                                                  n_steps=max_length,
                                                  outputs_info=outputs_info,
                                                  go_backwards=self.backwards,
                                                  non_sequences=non_sequence)
        return self.h_vals

    def get_initiation(self):
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0, T.constant(0.0, dtype=self.dtype)]

    def get_sequences(self):
        return [self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_ih, W_hh, b_h, projection, word_embedding):
        h_t = self.inner_activation(theano.dot(x_t, W_ih) + theano.dot(h_tm1, W_hh) + b_h)
        word_project = T.tanh(T.dot(h_t, projection))
        prediction_word_probability = T.nnet.softmax_graph(T.dot(word_embedding, word_project))
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    T.constant(0.0, dtype=self.dtype)], theano.scan_module.until(
                T.eq(predicted_word, self.stop_word_id))
        else:
            target_probability = prediction_word_probability[target_word]
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[target_word]
            return [target_word, output_word_embedding, h_t, loss]


class GRU(RNN):
    def __init__(self, gate_activation=T.nnet.sigmoid, **kwargs):
        # init parent attributes
        RNN.__init__(self, **kwargs)
        self.gate_activation = gate_activation
        self.W_iz = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_iz')
        self.W_hz = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hz')
        self.W_ir = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_ir')
        self.W_hr = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hr')
        self.b_z = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_z')
        self.b_r = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_r')
        self.params.extend([self.W_iz, self.W_hz, self.W_ir, self.W_hr, self.b_z, self.b_r])

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h]

    def one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h):
        zt = self.gate_activation(theano.dot(x_t, W_iz) + theano.dot(h_tm1, W_hz) + b_z)
        rt = self.gate_activation(theano.dot(x_t, W_ir) + theano.dot(h_tm1, W_hr) + b_r)
        rtht_1 = rt * h_tm1
        ht_hat = T.tanh(theano.dot(x_t, W_ih) + theano.dot(rtht_1, W_hh) + b_h)
        h_t = (1 - zt) * h_tm1 + zt * ht_hat
        return h_t


class LSTM(RNN):
    def __init__(self, **kwargs):
        # init parent attributes
        RNN.__init__(self, **kwargs)
        self.W_xi = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_xi')
        self.W_hi = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hi')
        self.W_ci = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_ci')
        self.b_i = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_i')
        self.W_xf = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_xf')
        self.W_hf = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hf')
        self.W_cf = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_cf')
        self.b_f = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_f')
        self.W_xc = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_xc')
        self.W_hc = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hc')
        self.b_c = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_c')
        self.W_xo = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_xo')
        self.W_ho = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_ho')
        self.W_co = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_co')
        self.b_o = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_o')
        self.c0 = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype))
        self.params = []
        self.params.extend(
            [self.W_xi, self.W_hi, self.W_ci, self.b_i, self.W_xf, self.W_hf, self.W_cf, self.b_f, self.W_xc,
             self.W_hc, self.b_c, self.W_xo, self.W_ho, self.W_co, self.b_o])

    def get_sequences(self):
        return [self.W_xi, self.W_hi, self.W_ci,
                self.b_i,
                self.W_xf,
                self.W_hf, self.W_cf, self.b_f,
                self.W_xc,
                self.W_hc,
                self.b_c, self.W_xo, self.W_ho, self.W_co, self.b_o]

    def get_initiation(self):
        if self.batch_mode:
            return [T.alloc(self.h0, self.sample_size, self.n_hidden),
                    T.alloc(self.c0, self.sample_size, self.n_hidden)]
        return [self.h0, self.c0]

    def one_step(self, x_t, h_tm1, c_tm1, W_xi, W_hi, W_ci, b_i, W_xf, W_hf, W_cf, b_f, W_xc, W_hc, b_c, W_xo,
                 W_ho, W_co, b_o):
        i_t = T.nnet.sigmoid(T.dot(x_t, W_xi) + T.dot(h_tm1, W_hi) + T.dot(c_tm1, W_ci) + b_i)
        f_t = T.nnet.sigmoid(T.dot(x_t, W_xf) + T.dot(h_tm1, W_hf) + T.dot(c_tm1, W_cf) + b_f)
        c_t = f_t * c_tm1 + i_t * self.inner_activation(T.dot(x_t, W_xc) + T.dot(h_tm1, W_hc) + b_c)
        o_t = T.nnet.sigmoid(T.dot(x_t, W_xo) + T.dot(h_tm1, W_ho) + T.dot(c_t, W_co) + b_o)
        h_t = o_t * self.inner_activation(c_t)
        return [h_t, c_t]

    @property
    def hidden_states(self):
        [self.h_vals, _] = self.h_vals
        if self.only_return_final:
            if self.return_method == 'ave':
                return T.mean(self.h_vals, axis=0)
            elif self.return_method == 'max':
                return T.max(self.h_vals, axis=0)
            else:
                return self.h_vals[-1]
        else:
            if self.backwards:
                self.h_vals = self.h_vals[::-1]
            if self.batch_mode:
                self.h_vals = self.h_vals.dimshuffle(1, 0, 2)
        return self.h_vals


class RUM(LSTM):
    def __init__(self, Lambda=1.0, eta=1.0, gate_activation=T.nnet.sigmoid, **kwargs):
        # init parent attributes
        LSTM.__init__(self, **kwargs)
        self.Lambda = Lambda
        self.eta = eta
        self.gate_activation = gate_activation
        self.params = []
        self.W_xh = self.init_one_parameter(shape=[self.n_in, 2 * self.n_hidden], name=self.name + 'W_xh')
        self.W_hh = self.init_one_parameter(shape=[self.n_hidden, 2 * self.n_hidden], name=self.name + 'W_hh')
        self.b_h = theano.shared(np.random.normal(0, 0.05, 2 * self.n_hidden).astype('float32'), name=self.name + 'b_h')
        self.W_th = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_th')
        self.b_t = theano.shared(np.random.normal(0, 0.05, self.n_hidden).astype('float32'), name=self.name + 'b_t')
        self.b_u = theano.shared(np.random.normal(0, 0.05, self.n_hidden).astype('float32'), name=self.name + 'b_u')
        self.r0 = theano.shared(np.eye(self.n_hidden, dtype='float32'))
        self.params.extend([self.W_xh, self.W_hh, self.W_th, self.b_h, self.b_t, self.b_u, self.r0])

    def get_sequences(self):
        return [self.W_xh, self.W_hh, self.b_h, self.W_th, self.b_t, self.b_u]

    def rotation(self, a, b):
        ua = a / (a.norm(L=2, axis=1).dimshuffle(0, 'x') + 1.0e-15)
        ub_ = (b - T.batched_dot(ua, b).dimshuffle(0, 'x') * ua)
        ub = ub_ / (ub_.norm(L=2, axis=1).dimshuffle(0, 'x') + 1.0e-15)
        cos = cosine(ua, b)
        sin = T.sqrt(1.0 - cos * cos)
        R = T.stack([cos, -sin, sin, cos]).T.reshape([sin.shape[0], 2, 2])
        uaa = ua.dimshuffle(0, 1, 'x') * ua.dimshuffle(0, 'x', 1)
        ubb = ub.dimshuffle(0, 1, 'x') * ub.dimshuffle(0, 'x', 1)
        uab = T.stack(ua, ub)
        rhs = T.batched_dot(T.batched_dot(uab.dimshuffle(1, 2, 0), R), uab.dimshuffle(1, 0, 2))
        lhs = T.eye(self.n_hidden, dtype='float32').dimshuffle('x', 0, 1) - uaa - ubb
        return rhs + lhs

    def get_initiation(self):
        return [T.alloc(self.h0, self.sample_size, self.n_hidden),
                T.alloc(self.r0, self.sample_size, self.n_hidden, self.n_hidden)]

    def modRelu(self, inputs, bt):
        inputs_ = inputs.norm(L=2, axis=1).dimshuffle(0, 'x') + 1.0e-15
        return (inputs / inputs_) * T.nnet.relu(inputs_ + bt.dimshuffle('x', 0))

    def one_step(self, x_t, h_tm1, rtm1, W_xh, W_hh, b_h, W_th, b_t, b_u):
        brac = T.dot(x_t, W_xh) + T.dot(h_tm1, W_hh) + b_h
        ut = T.nnet.sigmoid(brac[:, 0:self.n_hidden])
        tau = brac[:, self.n_hidden:]
        epsilong = T.dot(x_t, self.W_th) + b_t
        rt = (rtm1 ** self.Lambda) * self.rotation(epsilong, tau)
        _ht_ = (epsilong + T.batched_dot(rt, h_tm1))
        _ht = self.modRelu(_ht_, b_u)
        ht_ = ut * h_tm1 + (1.0 - ut) * _ht
        h = self.eta * ht_ / (ht_.norm(L=2, axis=1).dimshuffle(0, 'x') + 1.0e-15)
        return [h, rt]


class GRU_Decoder(GRU, RNN_Decoder):
    def __init__(self, **kwargs):
        # init parent attributes
        RNN_Decoder.__init__(self, **kwargs)
        GRU.__init__(self, **kwargs)

    def get_initiation(self):
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0, T.constant(0.0, dtype=self.dtype)]

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h, projection,
                 word_embedding):
        h_t = GRU.one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h)
        word_project = T.tanh(T.dot(h_t, projection))
        prediction_word_probability = T.nnet.softmax_graph(T.dot(word_embedding, word_project))
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    T.constant(0.0, dtype=self.dtype)], theano.scan_module.until(
                T.eq(predicted_word, self.stop_word_id))
        else:
            target_probability = prediction_word_probability[target_word]
            predicted_word = T.argmax(prediction_word_probability)
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[target_word]
            return [predicted_word, output_word_embedding, h_t, loss]


class RNN_attention_decoder(RNN_Decoder):
    def __init__(self,
                 attention_context=None,
                 **kwargs):
        RNN_Decoder.__init__(self, **kwargs)
        self.attention_context = attention_context
        self.attention_bilinear_matrix = self.init_one_parameter(shape=[self.n_encoder, self.n_hidden],
                                                                 name=self.name + 'attention_bilinear_matrix')
        self.attention_bilinear_bias = self.init_one_parameter(shape=[self.n_hidden],
                                                               name=self.name + 'attention_bilinear_bias')
        self.params.remove(self.projection_matrix_hidden)
        self.projection_matrix_hidden = self.init_one_parameter(shape=[self.n_encoder + self.n_hidden, self.n_in],
                                                                name=self.name + 'projection_matrix_hidden')
        self.params.append(self.projection_matrix_hidden)
        self.params.append(self.attention_bilinear_matrix)
        self.params.append(self.attention_bilinear_bias)

    def set_context(self, context):
        self.attention_context = context

    def get_sequences(self):
        return [self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_bilinear_matrix,
                self.attention_context]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_ih, W_hh, b_h, projection, word_embedding,
                 attention_bilinear_matrix, context):
        h_t = self.inner_activation(theano.dot(x_t, W_ih) + theano.dot(h_tm1, W_hh) + b_h)
        context_dot_value = T.dot(T.dot(context, attention_bilinear_matrix), h_t)
        context_softmax = T.nnet.softmax_graph(context_dot_value)
        context_representation = T.dot(context_softmax, context)
        hidden_representation = T.concatenate([context_representation, h_t])
        word_project = T.tanh(T.dot(hidden_representation, projection))
        prediction_word_probability = T.nnet.softmax_graph(T.dot(word_embedding, word_project))
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    T.constant(0.0, dtype=self.dtype)], theano.scan_module.until(
                T.eq(predicted_word, self.stop_word_id))
        else:
            target_probability = prediction_word_probability[target_word]
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[target_word]
            return [target_word, output_word_embedding, h_t, loss]


class GRU_attention_decoder(GRU, RNN_attention_decoder):
    def __init__(self,
                 vocab_size=0,
                 **kwargs):
        GRU.__init__(self, **kwargs)
        RNN_attention_decoder.__init__(self, **kwargs)
        self.vocab_size = vocab_size

    def get_initiation(self):
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0,
                theano.shared(np.zeros(self.n_in, dtype=self.dtype)),
                theano.shared(np.zeros(self.n_encoder, dtype=self.dtype)),
                theano.shared(np.zeros(self.vocab_size, dtype=self.dtype)),
                T.constant(0.0, dtype=self.dtype)]

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_bilinear_matrix, self.attention_bilinear_bias,
                self.attention_context]

    def one_step(self, target_word, __, x_t, h_tm1, ___, ____, _____, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh,
                 b_h,
                 projection,
                 word_embedding,
                 attention_bilinear_matrix, attention_bilinear_bias, context):
        h_t = GRU.one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h)
        context_to_hidden_representation = T.dot(context, attention_bilinear_matrix) + attention_bilinear_bias
        context_to_hidden_representation = T.tanh(context_to_hidden_representation)
        context_dot_value = T.tanh(T.dot(context_to_hidden_representation, h_t))
        context_softmax = T.nnet.softmax_graph(context_dot_value)
        context_representation = T.dot(context_softmax, context)
        hidden_representation = T.concatenate([context_representation, h_t])
        word_project = T.tanh(T.dot(hidden_representation, projection))
        word_distribution = T.dot(word_embedding, word_project)
        prediction_word_probability = T.nnet.softmax_graph(word_distribution)
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    word_project, context_representation, word_distribution,
                    T.constant(0.0, dtype=self.dtype)], theano.scan_module.until(
                T.eq(predicted_word, self.stop_word_id))
        else:
            target_probability = prediction_word_probability[target_word]
            predicted_word = T.argmax(prediction_word_probability)
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[target_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    word_project, context_representation, word_distribution, loss]


class GRU_attention_decoder_beta(GRU, RNN_attention_decoder):
    def __init__(self,
                 **kwargs):
        GRU.__init__(self, **kwargs)
        RNN_attention_decoder.__init__(self, **kwargs)
        self.C_r = self.init_one_parameter(shape=[self.n_encoder, self.n_hidden], name=self.name + 'C_r')
        self.C_z = self.init_one_parameter(shape=[self.n_encoder, self.n_hidden], name=self.name + 'C_u')
        self.C_h = self.init_one_parameter(shape=[self.n_encoder, self.n_hidden], name=self.name + 'C_h')
        self.params.remove(self.projection_matrix_hidden)
        self.projection_matrix_hidden = self.init_one_parameter(shape=[self.n_hidden, self.n_in],
                                                                name=self.name + 'projection_matrix_hidden')
        self.params.extend([self.C_r, self.C_z, self.C_h, self.projection_matrix_hidden])

    def get_initiation(self):
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0,
                theano.shared(np.zeros(self.n_in, dtype=self.dtype)),
                theano.shared(np.zeros(self.n_encoder, dtype=self.dtype)),
                theano.shared(np.zeros(50002, dtype=self.dtype)),
                T.constant(0.0, dtype=self.dtype)]

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_bilinear_matrix, self.attention_bilinear_bias,
                self.attention_context, self.C_r, self.C_z, self.C_h]

    def one_step(self, target_word, __, x_t, h_tm1, ___, ____, _____, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh,
                 b_h,
                 projection,
                 word_embedding,
                 attention_bilinear_matrix, attention_bilinear_bias, context, C_r, C_z, C_h):
        context_to_hidden_representation = T.dot(context, attention_bilinear_matrix) + attention_bilinear_bias
        context_to_hidden_representation = T.tanh(context_to_hidden_representation)
        context_dot_value = T.tanh(T.dot(context_to_hidden_representation, h_tm1))
        context_softmax = T.nnet.softmax_graph(context_dot_value)
        context_representation = T.dot(context_softmax, context)
        zt = self.gate_activation(
            theano.dot(x_t, W_iz) + theano.dot(h_tm1, W_hz) + T.dot(context_representation, C_z) + b_z)
        rt = self.gate_activation(
            theano.dot(x_t, W_ir) + theano.dot(h_tm1, W_hr) + T.dot(context_representation, C_r) + b_r)
        rtht_1 = rt * h_tm1
        ht_hat = T.tanh(theano.dot(x_t, W_ih) + theano.dot(rtht_1, W_hh) + T.dot(context_representation, C_h) + b_h)
        h_t = (1 - zt) * h_tm1 + zt * ht_hat
        word_project = T.tanh(T.dot(h_t, projection))
        word_distribution = T.dot(word_embedding, word_project)
        prediction_word_probability = T.nnet.softmax_graph(word_distribution)
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    word_project, context_representation, word_distribution,
                    T.constant(0.0, dtype=self.dtype)], theano.scan_module.until(
                T.eq(predicted_word, self.stop_word_id))
        else:
            target_probability = prediction_word_probability[target_word]
            predicted_word = T.argmax(prediction_word_probability)
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[target_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    word_project, context_representation, word_distribution, loss]


class GRU_context_attention_decoder(GRU_attention_decoder):
    def __init__(self,
                 context_score,
                 **kwargs):
        GRU_attention_decoder.__init__(self, **kwargs)
        self.context_score = context_score

    def get_initiation(self):
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0, T.constant(0.0, dtype=self.dtype)]

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_bilinear_matrix,
                self.attention_context, self.context_score]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h, projection,
                 word_embedding,
                 attention_bilinear_matrix, context, context_score):
        h_t = GRU.one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h)
        context_dot_value = T.dot(T.dot(context, attention_bilinear_matrix), h_t)
        context_softmax = T.nnet.softmax_graph(context_score * context_dot_value)
        context_representation = T.dot(context_softmax, context)
        hidden_representation = T.concatenate([context_representation, h_t])
        word_project = T.tanh(T.dot(hidden_representation, projection))
        prediction_word_probability = T.nnet.softmax_graph(T.dot(word_embedding, word_project))
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    T.constant(0.0, dtype=self.dtype)], theano.scan_module.until(
                T.eq(predicted_word, self.stop_word_id))
        else:
            target_probability = prediction_word_probability[target_word]
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[target_word]
            return [target_word, output_word_embedding, h_t, loss]


class Batched_RNN_Decoder(RNN):
    def __init__(self,
                 e_d_shape_match,
                 word_embedding,
                 start_hidden_size=None,
                 n_encoder=None,
                 **kwargs):
        RNN.__init__(self, **kwargs)
        self.e_d_shape_match = e_d_shape_match
        self.n_encoder = n_encoder
        if start_hidden_size is None:
            start_hidden_size = self.n_encoder
        if e_d_shape_match is not True:
            if n_encoder is None:
                assert 'you must specify the encoder dimension'
            else:
                self.en_de_projection = self.init_one_parameter(shape=[start_hidden_size, self.n_hidden],
                                                                name=self.name + 'en_de_projection')
                self.params.append(self.en_de_projection)

        self.projection_matrix_hidden = self.init_one_parameter(
            shape=[self.n_hidden + self.n_encoder, self.n_in],
            name=self.name + 'projection_matrix_hidden')
        self.params.append(self.projection_matrix_hidden)
        self.mode = None
        self.start_word = T.constant(-1, dtype='int32')
        self.start_word_embedding = theano.shared(np.zeros(self.n_in, dtype=self.dtype),
                                                  name=self.name + '_x0_')
        self.word_embedding = word_embedding
        self.stop_word_id = word_embedding.shape[0] - 1

    def decode(self, max_length, start_hidden=None, true_word=None):
        if self.e_d_shape_match:
            self.set_h0(start_hidden)
        else:
            encoder_input = T.tanh(T.dot(start_hidden, self.en_de_projection))
            self.set_h0(encoder_input)
        if true_word is not None and max_length is None:
            self.mode = 'Train'
            max_length = None
        elif true_word is None and max_length is not None:
            self.mode = 'Test'
            true_word = T.arange(0, max_length, dtype='int32')
        else:
            self.mode = 'Evaluation'
            max_length = None
        if self.batch_mode:
            self.sample_size = start_hidden.shape[0]
        print self.mode
        outputs_info = self.get_initiation()
        non_sequence = self.get_sequences()
        step_fun = self.one_step
        self.h_vals = None
        self.h_vals, self.__updates = theano.scan(fn=step_fun,
                                                  sequences=true_word,
                                                  n_steps=max_length,
                                                  outputs_info=outputs_info,
                                                  go_backwards=self.backwards,
                                                  non_sequences=non_sequence)
        return self.h_vals

    def get_initiation(self):
        if self.batch_mode:
            return [T.zeros(shape=[self.sample_size], dtype='int32'),
                    T.alloc(self.start_word_embedding, self.sample_size, self.n_in),
                    self.h0,
                    T.zeros(shape=[self.sample_size], dtype='float32'),
                    ]
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0,
                T.constant(0.0, dtype=self.dtype)]

    def get_sequences(self):
        return [self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_ih, W_hh, b_h, projection, word_embedding):
        h_t = self.inner_activation(theano.dot(x_t, W_ih) + theano.dot(h_tm1, W_hh) + b_h)
        word_project = T.tanh(T.dot(h_t, projection))
        prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
        return self.return_value(prediction_word_probability, word_embedding, h_t, target_word)

    def return_value(self, prediction_word_probability, word_embedding, h_t, target_word):
        if self.mode == 'Test':
            predicted_word = T.argmax(prediction_word_probability, axis=1)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    T.zeros(shape=[self.sample_size], dtype='float32')]
        elif self.mode == 'Train':
            # target_probability = T.diag(prediction_word_probability.T[target_word])  # this is faster in practice
            # loss = -T.log(target_probability)
            # output_word_embedding = word_embedding[target_word]
            # return [target_word, output_word_embedding, h_t, loss]
            predicted_word = T.argmax(prediction_word_probability, axis=1)
            target_probability = T.diag(
                prediction_word_probability.T[target_word])  # this is faster in practice
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t, loss]
        else:
            predicted_word = T.argmax(prediction_word_probability, axis=1)
            target_probability = T.diag(
                prediction_word_probability.T[target_word])  # this is faster in practice
            loss = -T.log(target_probability)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t, loss]


class Batched_ATTENTION_RNN_Decoder(Batched_RNN_Decoder):
    """
    For the encoder that do not have the length variance
    In our case it is the length of the context
    """

    def __init__(self,
                 attention_source,
                 recoder_size=None,
                 **kwargs):
        Batched_RNN_Decoder.__init__(self, **kwargs)
        if recoder_size is None:
            self.recoder_size = self.n_encoder
        self.attention_source = attention_source
        self.attention_bilinear_matrix = self.init_one_parameter(shape=[self.n_encoder, self.n_hidden],
                                                                 name=self.name + 'attention_bilinear_matrix')
        self.params.remove(self.projection_matrix_hidden)
        self.projection_matrix_hidden = self.init_one_parameter(
            shape=[self.n_encoder + self.n_hidden, self.n_in],
            name=self.name + 'projection_matrix_hidden')
        self.params.append(self.attention_bilinear_matrix)
        self.params.append(self.projection_matrix_hidden)

    def set_attention_source(self, attention_source):
        self.attention_source = attention_source

    def get_sequences(self):
        transformed_context_representation = T.tanh(
            T.dot(self.attention_source, self.attention_bilinear_matrix))  # batch*max_len*hidden
        return [self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_source, transformed_context_representation]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_ih, W_hh, b_h, projection, word_embedding,
                 attention_source,
                 transoformed_context_representation):
        h_t = self.inner_activation(theano.dot(x_t, W_ih) + theano.dot(h_tm1, W_hh) + b_h)
        context_dot_value = T.tanh(T.batched_dot(transoformed_context_representation, h_t))
        context_softmax = T.nnet.softmax(context_dot_value)
        context_attentive_representation = T.batched_dot(attention_source.dimshuffle(0, 2, 1), context_softmax)
        hidden_representation = T.concatenate([context_attentive_representation, h_t], axis=1)
        word_project = T.tanh(T.dot(hidden_representation, projection))
        prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
        return self.return_value(prediction_word_probability, word_embedding, h_t, target_word)


class Batched_ATTENTION_GRU_Decoder(Batched_ATTENTION_RNN_Decoder, GRU):
    def __init__(self,
                 **kwargs):
        Batched_ATTENTION_RNN_Decoder.__init__(self, **kwargs)
        GRU.__init__(self, **kwargs)

    def get_sequences(self):
        transformed_context_representation = T.tanh(
            T.dot(self.attention_source, self.attention_bilinear_matrix))  # batch*max_len*hidden
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_source, transformed_context_representation]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h,
                 projection,
                 word_embedding, attention_source,
                 transoformed_context_representation):
        h_t = GRU.one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h)
        context_dot_value = T.tanh(T.batched_dot(transoformed_context_representation, h_t))
        context_softmax = T.nnet.softmax(context_dot_value)
        context_attentive_representation = T.batched_dot(context_softmax, attention_source)
        hidden_representation = T.concatenate([context_attentive_representation, h_t], axis=1)
        word_project = T.tanh(T.dot(hidden_representation, projection))
        prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
        return self.return_value(prediction_word_probability, word_embedding, h_t, target_word)


class Batched_ATTENTION_LSTM_Decoder(Batched_ATTENTION_RNN_Decoder, GRU):
    def __init__(self,
                 **kwargs):
        Batched_ATTENTION_RNN_Decoder.__init__(self, **kwargs)
        GRU.__init__(self, **kwargs)

    def get_sequences(self):
        transformed_context_representation = T.tanh(
            T.dot(self.attention_source, self.attention_bilinear_matrix))  # batch*max_len*hidden
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.attention_source, transformed_context_representation]

    def one_step(self, target_word, __, x_t, h_tm1, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h,
                 projection,
                 word_embedding, attention_source,
                 transoformed_context_representation):
        h_t = GRU.one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h)
        context_dot_value = T.tanh(T.batched_dot(transoformed_context_representation, h_t))
        context_softmax = T.nnet.softmax(context_dot_value)
        context_attentive_representation = T.batched_dot(context_softmax, attention_source)
        hidden_representation = T.concatenate([context_attentive_representation, h_t], axis=1)
        word_project = T.tanh(T.dot(hidden_representation, projection))
        prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
        return self.return_value(prediction_word_probability, word_embedding, h_t, target_word)


class Batched_GRU_Decoder(Batched_RNN_Decoder, GRU):
    def __init__(self,
                 encoder_representation=None,
                 **kwargs):
        Batched_RNN_Decoder.__init__(self, **kwargs)
        GRU.__init__(self, **kwargs)
        if encoder_representation is not None:
            self.encoder_representation = encoder_representation

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.projection_matrix_hidden, self.word_embedding,
                self.encoder_representation]

    def set_mode(self, mode):
        self.mode = mode

    def one_step(self, target_word, __, x_t, h_tm1, _, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h,
                 projection,
                 word_embedding, encoder_projection):
        h_t = GRU.one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h)
        hidden = T.concatenate([h_t, encoder_projection], axis=-1)
        word_project = T.tanh(T.dot(hidden, projection))
        prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
        return self.return_value(prediction_word_probability, word_embedding, h_t, target_word)


class Batched_LSTM_Decoder(Batched_RNN_Decoder, LSTM):
    def __init__(self,
                 encoder_representation,
                 **kwargs):
        LSTM.__init__(self, **kwargs)
        Batched_RNN_Decoder.__init__(self, **kwargs)
        if encoder_representation is not None:
            self.encoder_representation = encoder_representation

    def set_encoder_representation(self, encoder_representation):
        self.encoder_representation = encoder_representation

    def get_sequences(self):
        return [self.W_xi, self.W_hi, self.W_ci,
                self.b_i,
                self.W_xf,
                self.W_hf, self.W_cf, self.b_f,
                self.W_xc,
                self.W_hc,
                self.b_c, self.W_xo, self.W_ho, self.W_co, self.b_o, self.projection_matrix_hidden,
                self.word_embedding, self.encoder_representation]

    def set_mode(self, mode):
        self.mode = mode

    def get_initiation(self):
        if self.batch_mode:
            return [T.zeros(shape=[self.sample_size], dtype='int32'),
                    T.alloc(self.start_word_embedding, self.sample_size, self.n_in),
                    self.h0,
                    T.alloc(theano.shared(np.zeros(self.n_hidden, dtype=self.dtype)), self.sample_size,
                            self.n_hidden),
                    T.zeros(shape=[self.sample_size], dtype='float32'),
                    ]
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0, self.c0,
                T.constant(0.0, dtype=self.dtype)]

    def one_step(self, target_word, __, x_t, h_tm1, c_tm1, _, W_xi, W_hi, W_ci, b_i, W_xf, W_hf, W_cf, b_f, W_xc,
                 W_hc,
                 b_c, W_xo,
                 W_ho, W_co, b_o,
                 projection,
                 word_embedding, encoder_representation):
        [h_t, c_t] = LSTM.one_step(self, x_t, h_tm1, c_tm1, W_xi, W_hi, W_ci, b_i, W_xf, W_hf, W_cf, b_f, W_xc,
                                   W_hc,
                                   b_c, W_xo,
                                   W_ho, W_co, b_o)
        hidden = T.concatenate([h_t, encoder_representation], axis=-1)
        word_project = T.tanh(T.dot(hidden, projection))
        prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
        [target_word, output_word_embedding, h_t, loss] = self.return_value(prediction_word_probability,
                                                                            word_embedding,
                                                                            h_t, target_word)
        return [target_word, output_word_embedding, h_t, c_t, loss]


class Batched_LSTM_Decoder_IS(LSTM):
    def __init__(self,
                 encoder_representation,
                 fake_word_embedding,
                 word_embedding,
                 n_encoder,
                 **kwargs):
        LSTM.__init__(self, **kwargs)
        self.n_encoder = n_encoder
        self.word_embedding = word_embedding
        self.fake_word_embedding = fake_word_embedding
        if encoder_representation is not None:
            self.encoder_representation = encoder_representation
        start_hidden_size = self.n_encoder
        self.en_de_projection = self.init_one_parameter(shape=[start_hidden_size, self.n_hidden],
                                                        name=self.name + 'en_de_projection')
        self.params.append(self.en_de_projection)

        self.projection_matrix_hidden = self.init_one_parameter(shape=[self.n_hidden + self.n_encoder, self.n_in],
                                                                name=self.name + 'projection_matrix_hidden')
        self.params.append(self.projection_matrix_hidden)
        self.mode = None
        self.start_word = T.constant(-1, dtype='int32')
        self.start_word_embedding = theano.shared(np.zeros(self.n_in, dtype=self.dtype), name=self.name + '_x0_')

    def set_encoder_representation(self, encoder_representation):
        self.encoder_representation = encoder_representation

    def get_sequences(self):
        return [self.W_xi, self.W_hi, self.W_ci,
                self.b_i,
                self.W_xf,
                self.W_hf, self.W_cf, self.b_f,
                self.W_xc,
                self.W_hc,
                self.b_c, self.W_xo, self.W_ho, self.W_co, self.b_o, self.projection_matrix_hidden,
                self.word_embedding, self.encoder_representation]

    def set_mode(self, mode):
        self.mode = mode

    def get_initiation(self):
        if self.batch_mode:
            return [T.zeros(shape=[self.sample_size], dtype='int32'),
                    T.alloc(self.start_word_embedding, self.sample_size, self.n_in),
                    self.h0,
                    T.alloc(theano.shared(np.zeros(self.n_hidden, dtype=self.dtype)), self.sample_size,
                            self.n_hidden),
                    T.zeros(shape=[self.sample_size], dtype='float32'),
                    ]
        return [T.constant(0, dtype='int32'), self.start_word_embedding, self.h0, self.c0,
                T.constant(0.0, dtype=self.dtype)]

    def decode(self, max_length, start_hidden=None, true_word=None):
        encoder_input = T.tanh(T.dot(start_hidden, self.en_de_projection))
        self.set_h0(encoder_input)
        if true_word is not None and max_length is None:
            self.mode = 'Train'
            max_length = None
        elif true_word is None and max_length is not None:
            self.mode = 'Test'
            true_word = T.arange(0, max_length, dtype='int32')
        else:
            self.mode = 'Evaluation'
            max_length = None
        if self.batch_mode:
            self.sample_size = start_hidden.shape[0]
        print self.mode
        outputs_info = self.get_initiation()
        non_sequence = self.get_sequences()
        step_fun = self.one_step
        self.h_vals = None
        self.h_vals, self.__updates = theano.scan(fn=step_fun,
                                                  sequences=[true_word, self.fake_word_embedding],
                                                  n_steps=max_length,
                                                  outputs_info=outputs_info,
                                                  go_backwards=self.backwards,
                                                  non_sequences=non_sequence)
        return self.h_vals

    def one_step(self, target_word, fake_word_embedding, __, x_t, h_tm1, c_tm1, _, W_xi, W_hi, W_ci, b_i, W_xf, W_hf,
                 W_cf, b_f,
                 W_xc,
                 W_hc,
                 b_c, W_xo,
                 W_ho, W_co, b_o,
                 projection,
                 word_embedding, encoder_representation):
        [h_t, c_t] = LSTM.one_step(self, x_t, h_tm1, c_tm1, W_xi, W_hi, W_ci, b_i, W_xf, W_hf, W_cf, b_f, W_xc,
                                   W_hc,
                                   b_c, W_xo,
                                   W_ho, W_co, b_o)
        hidden = T.concatenate([h_t, encoder_representation], axis=-1)
        word_project = T.tanh(T.dot(hidden, projection))
        if self.mode == 'Test':
            prediction_word_probability = T.nnet.softmax(T.dot(word_project, word_embedding.T))
            predicted_word = T.argmax(prediction_word_probability, axis=1)
            output_word_embedding = word_embedding[predicted_word]
            return [T.cast(predicted_word, 'int32'), output_word_embedding, h_t,
                    T.zeros(shape=[self.sample_size], dtype='float32')]
        output_word_embedding = word_embedding[target_word]
        bucket_word_embedding = T.concatenate(
            [output_word_embedding.reshape([output_word_embedding.shape[0], 1, output_word_embedding.shape[1]]),
             fake_word_embedding], axis=1)
        prediction_word_probability = T.nnet.softmax(T.batched_dot(bucket_word_embedding, word_project))
        loss = -T.log(prediction_word_probability[:, 0])
        return [target_word, output_word_embedding, h_t, c_t, loss]


class MultiLayerRNN(object):
    def __init__(self, n_word_embedding, n_hidden, n_layer=3, name='', cell=LSTM):
        self.n_layer = n_layer
        self.encoder_forward = []
        self.encoder_backward = []
        self.parameter = []
        for i in range(self.n_layer):
            if i == 0:
                n_in = n_word_embedding
            else:
                n_in = 2 * n_hidden
            encoder_forward = cell(n_in=n_in, n_hidden=n_hidden, only_return_final=False,
                                   name=name + '_encoder_forward' + str(i))
            encoder_backward = cell(n_in=n_in, n_hidden=n_hidden, only_return_final=False,
                                    backwards=True,
                                    name=name + '_encoder_backward_' + str(i))
            self.parameter.extend(encoder_forward.parameter)
            self.parameter.extend(encoder_backward.parameter)
            self.encoder_forward.append(encoder_forward)
            self.encoder_backward.append(encoder_backward)

    def __call__(self, word_representation):
        context_word_embedding = word_representation
        for i in range(self.n_layer):
            encoder_hidden_forward = self.encoder_forward[i](context_word_embedding, batch_mode=True)
            encoder_hidden_backward = self.encoder_backward[i](context_word_embedding, batch_mode=True)
            word_representation = T.concatenate([encoder_hidden_forward, encoder_hidden_backward], axis=-1)
            context_word_embedding = word_representation
        return context_word_embedding


class MultiLayerGRUCONTEXT(object):
    def __init__(self, attention_representation, n_attention, n_word_embedding, n_hidden, n_layer=3, name=''):
        self.attention_representation = attention_representation
        self.n_layer = n_layer
        self.encoder_forward = []
        self.encoder_backward = []
        self.parameter = []
        for i in range(self.n_layer):
            if i == 0:
                n_in = n_word_embedding
            else:
                n_in = 2 * n_hidden
            encoder_forward = GRU_Context(n_in=n_in, attention_representation=attention_representation,
                                          n_attention=n_attention, n_hidden=n_hidden,
                                          only_return_final=False,
                                          name=name + '_encoder_forward' + str(i))
            encoder_backward = GRU_Context(n_in=n_in, attention_representation=attention_representation,
                                           n_attention=n_attention, n_hidden=n_hidden,
                                           only_return_final=False,
                                           backwards=True,
                                           name=name + '_encoder_backward_' + str(i))
            self.parameter.extend(encoder_forward.parameter)
            self.parameter.extend(encoder_backward.parameter)
            self.encoder_forward.append(encoder_forward)
            self.encoder_backward.append(encoder_backward)

    def __call__(self, word_representation):
        context_word_embedding = word_representation
        for i in range(self.n_layer):
            encoder_hidden_forward = self.encoder_forward[i](context_word_embedding, batch_mode=True)
            encoder_hidden_backward = self.encoder_backward[i](context_word_embedding, batch_mode=True)
            word_representation = T.concatenate([encoder_hidden_forward, encoder_hidden_backward], axis=-1)
            context_word_embedding = word_representation
        return context_word_embedding


class MultiLayerGRUGATE(object):
    def __init__(self, attention_representation, n_attention, n_word_embedding, n_hidden, n_layer=3, name=''):
        self.attention_representation = attention_representation
        self.n_layer = n_layer
        self.encoder_forward = []
        self.encoder_backward = []
        self.parameter = []
        for i in range(self.n_layer):
            if i == 0:
                n_in = n_word_embedding
            else:
                n_in = 2 * n_hidden
            encoder_forward = GRU_Gate(n_in=n_in, attention_representation=attention_representation,
                                       n_attention=n_attention, n_hidden=n_hidden,
                                       only_return_final=False,
                                       name=name + '_encoder_forward' + str(i))
            encoder_backward = GRU_Gate(n_in=n_in, attention_representation=attention_representation,
                                        n_attention=n_attention, n_hidden=n_hidden,
                                        only_return_final=False,
                                        backwards=True,
                                        name=name + '_encoder_backward_' + str(i))
            self.parameter.extend(encoder_forward.parameter)
            self.parameter.extend(encoder_backward.parameter)
            self.encoder_forward.append(encoder_forward)
            self.encoder_backward.append(encoder_backward)

    def __call__(self, word_representation):
        context_word_embedding = word_representation
        for i in range(self.n_layer):
            encoder_hidden_forward = self.encoder_forward[i](context_word_embedding, batch_mode=True)
            encoder_hidden_backward = self.encoder_backward[i](context_word_embedding, batch_mode=True)
            word_representation = T.concatenate([encoder_hidden_forward, encoder_hidden_backward], axis=-1)
            context_word_embedding = word_representation
        return context_word_embedding


class GRU_Context(RNN):
    def __init__(self, attention_representation, n_attention, gate_activation=T.nnet.sigmoid, **kwargs):
        # init parent attributes
        RNN.__init__(self, **kwargs)
        self.n_attention = n_attention
        self.gate_activation = gate_activation
        self.W_iz = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_iz')
        self.W_hz = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hz')
        self.W_ir = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_ir')
        self.W_hr = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hr')
        self.b_z = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_z')
        self.b_r = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_r')
        self.attention_representation_project = self.init_one_parameter(shape=[self.n_attention, self.n_in],
                                                                        name=self.name + 'W_a_r_p')
        self.context_representation_project = self.init_one_parameter(shape=[self.n_hidden, self.n_in],
                                                                      name=self.name + 'W_a_r_p')

        self.params.extend(
            [self.W_iz, self.W_hz, self.W_ir, self.W_hr, self.b_z, self.b_r, self.attention_representation_project,
             self.context_representation_project])
        self.attention_representation = T.dot(attention_representation, self.attention_representation_project)

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h, self.context_representation_project,
                self.attention_representation]

    def one_step(self, x_t_, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h, context_representation_project,
                 attention_representation_):
        attention_representation = T.dot(h_tm1, context_representation_project) + attention_representation_
        attention_source = T.nnet.sigmoid(T.batched_dot(attention_representation, x_t_))
        x_t = attention_source.dimshuffle(0, 'x') * x_t_
        zt = self.gate_activation(theano.dot(x_t, W_iz) + theano.dot(h_tm1, W_hz) + b_z)
        rt = self.gate_activation(theano.dot(x_t, W_ir) + theano.dot(h_tm1, W_hr) + b_r)
        rtht_1 = rt * h_tm1
        ht_hat = T.tanh(theano.dot(x_t, W_ih) + theano.dot(rtht_1, W_hh) + b_h)
        h_t = (1 - zt) * h_tm1 + zt * ht_hat
        return h_t


class GRU_Gate(RNN):
    def __init__(self, attention_representation, n_attention, gate_activation=T.nnet.sigmoid, **kwargs):
        # init parent attributes
        RNN.__init__(self, **kwargs)
        self.n_attention = n_attention
        self.gate_activation = gate_activation
        self.W_iz = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_iz')
        self.W_hz = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hz')
        self.W_ir = self.init_one_parameter(shape=[self.n_in, self.n_hidden], name=self.name + 'W_ir')
        self.W_hr = self.init_one_parameter(shape=[self.n_hidden, self.n_hidden], name=self.name + 'W_hr')
        self.b_z = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_z')
        self.b_r = theano.shared(np.zeros(self.n_hidden, dtype=self.dtype), name=self.name + 'b_r')
        self.attention_representation_project = self.init_one_parameter(shape=[self.n_attention, self.n_hidden],
                                                                        name=self.name + 'W_a_r_p')

        self.params.extend(
            [self.W_iz, self.W_hz, self.W_ir, self.W_hr, self.b_z, self.b_r, self.attention_representation_project])
        self.attention_representation = T.dot(attention_representation, self.attention_representation_project)

    def get_sequences(self):
        return [self.W_iz, self.W_hz, self.b_z, self.W_ir, self.W_hr,
                self.b_r, self.W_ih, self.W_hh, self.b_h,
                self.attention_representation]

    def one_step(self, x_t, h_tm1, W_iz, W_hz, b_z, W_ir, W_hr, b_r, W_ih, W_hh, b_h,
                 attention_representation_):
        zt = self.gate_activation(theano.dot(x_t, W_iz) + theano.dot(h_tm1, W_hz) + attention_representation_ + b_z)
        rt = self.gate_activation(theano.dot(x_t, W_ir) + theano.dot(h_tm1, W_hr) + attention_representation_ + b_r)
        rtht_1 = rt * h_tm1
        ht_hat = T.tanh(theano.dot(x_t, W_ih) + theano.dot(rtht_1, W_hh) + b_h)
        h_t = (1 - zt) * h_tm1 + zt * ht_hat
        return h_t

#
# if __name__ == '__main__':
#     batch_size = 21
#     word_number = 500
#     sentence_length = 35
#     embedding_size = 100
#     encoder_size = 150
#     decoder_size = 260
#     word_embedding = rng.normal(size=(word_number, embedding_size))
#     Word_embedding = theano.shared(word_embedding, name='word_embedding')
#     inputs_forward = T.imatrix('input_document_forward')  # 21*max_length
#     inputs_backward = T.imatrix('input_document_backward')  # 21*max_length
#     input_length = T.ivector('input_length')
#     rnn_forward = GRU(n_in=embedding_size, n_hidden=encoder_size, only_return_final=False)
#     rnn_backward = GRU(n_in=embedding_size, n_hidden=encoder_size, only_return_final=False,
#                        name='encoder_backward')  # we have already reverse the list
#     Inputs_forward = Word_embedding[inputs_forward]
#     Inputs_backward = Word_embedding[inputs_forward]
#     encoder_hidden_forward = rnn_forward(Inputs_forward, batch_mode=True)
#     encoder_hidden_backward = rnn_backward(Inputs_backward, batch_mode=True)
#     output_vector = T.concatenate([encoder_hidden_forward, encoder_hidden_backward], axis=2)
#     hidden_representation = []
#     for i in xrange(20):
#         hidden_representation.append([output_vector[i][input_length[i] - 1]])
#     context_representation = T.concatenate(hidden_representation, axis=0)
#     target_representation = output_vector[20][input_length[20] - 1]
#     context_to_target_representation = T.mean(context_representation, axis=0)
#     context_score = T.nnet.sigmoid(T.dot(context_representation, target_representation))
#     decoder = GRU_context_attention_decoder(e_d_shape_match=False, word_embedding=Word_embedding,
#                                             n_encoder=encoder_size * 2,
#                                             attention_context=context_representation,
#                                             context_score=context_score,
#                                             n_hidden=decoder_size, n_in=embedding_size)
#     decoder_stuff = decoder.decode(start_hidden=context_to_target_representation, max_length=None,
#                                    true_word=inputs_forward[20][0:input_length[-1]])
#     test = theano.function([inputs_forward, inputs_backward, input_length], outputs=[decoder_stuff[-1],context_to_target_representation,target_representation],
#                            on_unused_input='ignore',
#                            allow_input_downcast=True)
#
#     for i in range(200000):
#         cc_forward = [np.random.randint(1, word_number - 1, size=np.random.randint(5, sentence_length)) for _ in
#                       xrange(batch_size)]
#         cc_backward = [x[::-1] for x in cc_forward]
#         cc_forward, L1 = padding(cc_forward)
#         cc_backward, _ = padding(cc_backward)
#         output = test(cc_forward, cc_backward, L1)
#         print i
