# -*- coding:utf-8 -*-
# Author: Bingning Wang (王炳宁)
# Email: research@bingning.wang
# HomePage: www.bingning.wang
# Date: 15-10-15 下午1:05

import cPickle
import time
import theano.tensor as T
import theano
import numpy as np
import lasagne

rng = np.random.RandomState(2016)
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

irng = RandomStreams(2009)


def parameter_size(parameters):
    p_size = {}
    for parameter in parameters:
        name = parameter.name
        value = parameter.get_value(borrow=True)
        size = value.shape
        p_size[name] = size
    total = 0
    for one in p_size.values():
        a_size = 1
        for i in one:
            a_size *= i
        total += a_size
    return p_size, total


def Xavier_init(shape, name):
    n_ = np.sqrt(2.0 / sum(shape))
    w_ = np.random.normal(0, 1, shape) * n_
    return theano.shared(np.float32(w_), name=name)


def get_function(inputs, outputs, updates=None):
    f = theano.function(inputs=inputs,
                        outputs=outputs,
                        updates=updates,
                        allow_input_downcast=True,
                        on_unused_input='ignore')
    return f


def NormInit(shape, name):
    scale = 1.0 / shape[0]
    return theano.shared(np.float32(np.random.normal(0, scale, shape)), name=name)


def softmax_3d(inputs):
    exp = T.exp(inputs)
    sum_value = T.sum(exp, axis=-1).dimshuffle(0, 1, 'x')
    return exp / sum_value


def init_parameter(sizeX, sizeY, name=''):
    return theano.shared(sample_weights(sizeX, sizeY), name=name)


def init_bias(sizeX, name=''):
    return theano.shared(np.float32(np.random.normal(0, 0.1, size=sizeX)), name=name)


def get_update(optmizer, loss, parameter, learning_rate=0.01):
    if optmizer == 'sgd':
        updates = lasagne.updates.sgd(loss, parameter, learning_rate=learning_rate)
    elif optmizer == 'adadelta':
        updates = lasagne.updates.adadelta(loss, parameter)
    elif optmizer == 'adagrad':
        updates = lasagne.updates.adagrad(loss, parameter)
    elif optmizer == 'rmsprop':
        updates = lasagne.updates.rmsprop(loss, parameter)
    elif optmizer == 'adam':
        updates = lasagne.updates.adam(loss_or_grads=loss, params=parameter)
    elif optmizer == 'momentum':
        updates = lasagne.updates.momentum(loss_or_grads=loss, params=parameter, learning_rate=learning_rate)
    elif optmizer == 'nesterov_momentum':
        updates = lasagne.updates.nesterov_momentum(loss_or_grads=loss, params=parameter, learning_rate=learning_rate)
    elif optmizer == 'adamax':
        updates = lasagne.updates.adamax(loss_or_grads=loss, params=parameter)
    else:
        updates = lasagne.updates.sgd(loss, parameter, learning_rate=learning_rate)
    return updates


def dropout_from_layer(layer, p):
    """p is the probablity of dropping a unit
    """
    # p=1-p because 1's indicate keep and p is prob of dropping
    mask = irng.binomial(n=1, p=1 - p, size=layer.shape, dtype='float32')
    output = layer * mask
    return output


def SELU(x):
    alpha = 1.6732632423543772848170429916717
    scale = 1.0507009873554804934193349852946
    return scale * T.switch(x > 0, x, alpha * (T.exp(x) - 1))


def clip_value(tensor):
    return T.clip(tensor, 1.0e-15, 1.0 - 1.0e-15)


def clip_softmax(input_vector):
    max_value = T.max(abs(input_vector))
    input_vector /= max_value
    return T.nnet.softmax_graph(input_vector)


def cosine_distance(vector, matrix):
    dominator = T.dot(matrix, vector)
    numerator = vector.norm(2, axis=0) * matrix.norm(2, axis=1)
    return dominator / numerator


def add_l1_l2_norm(l1, l2, loss, input_parameter):
    for param in input_parameter:
        loss += l1 * T.sum(abs(param)) + l2 * T.sum(param ** 2)
    return loss


def my_softmax(x):
    prob = x - T.max(x, axis=-1, keepdims=True)
    lprob = prob - T.log(T.sum(T.exp(prob), axis=-1, keepdims=True))
    return T.exp(lprob)


def norm(parameter, l1=False):
    if l1:
        return T.sum(abs(parameter))
    return T.sqrt(T.sum(parameter ** 2))


def matrix_vector_cosine_similarity(matrix, vecotr):
    num = T.dot(matrix, vecotr)
    norm1 = matrix.norm(2, axis=1)
    norm2 = vecotr.norm(2)
    return num / ((norm1 * norm2) + 1.0e-10)


def sample_weights(sizeX, sizeY, low=-1., high=1., method='svd'):
    """
        it has been proved that the max singular value of a matirx can not
        exceed 1 for the non exploding RNN issues
        :param high: high bound
        :param low: low bound
        :param sizeY: the initiation matrix size y
        :param sizeX:the initiation matrix size x
        :return: the svd matrix remove max value
        """
    if method == 'random':
        return rng.normal(size=(sizeX, sizeY), loc=0.5, scale=0.3)
    else:
        values = np.ndarray([sizeX, sizeY], dtype=theano.config.floatX)
        for dx in xrange(sizeX):
            vals = np.random.uniform(low=low, high=high, size=(sizeY,))
            # vals_norm = np.sqrt((vals**2).sum())
            # vals = vals / vals_norm
            values[dx, :] = vals
        _, svs, _ = np.linalg.svd(values)
        # svs[0] is the largest singular value
        values /= 2 * svs[0]
        return values


def cosine(x, y):
    return T.dot(x, y) / (T.sqrt(T.sum(x ** 2)) * T.sqrt(T.sum(y ** 2)))


#
# def cosine_tf(x, y, length):
#     return dot_vectors(x, y, length) / (tensorflow_l2_norm(x) * tensorflow_l2_norm(y))
#

def batch_cosine(x, y):
    num = T.batched_dot(x, y)
    dom = x.norm(L=2, axis=1) * y.norm(L=2, axis=1)
    return num / dom


def nd_softmax(a, axis=-1):
    swapped_a = a.swapaxes(axis, -1)
    matrix_a = swapped_a.reshape((swapped_a.shape[:-1].prod(), a.shape[-1]))
    softmax_swapped_a = T.nnet.softmax(matrix_a).reshape(swapped_a.shape)
    return softmax_swapped_a.swapaxes(axis, -1)


def log_sum_exp(x, axis=None):
    """
    Sum probabilities in the log-space.
    """
    xmax = x.max(axis=axis, keepdims=True)
    xmax_ = x.max(axis=axis)
    return xmax_ + T.log(T.exp(x - xmax).sum(axis=axis))


def forward(observations, transitions, viterbi=False,
            return_alpha=False, return_best_sequence=False):
    """
    Takes as input:
        - observations, sequence of shape (n_steps, n_classes)
        - transitions, sequence of shape (n_classes, n_classes)
    Probabilities must be given in the log space.
    Compute alpha, matrix of size (n_steps, n_classes), such that
    alpha[i, j] represents one of these 2 values:
        - the probability that the real path at node i ends in j
        - the maximum probability of a path finishing in j at node i (Viterbi)
    Returns one of these 2 values:
        - alpha
        - the final probability, which can be:
            - the sum of the probabilities of all paths
            - the probability of the best path (Viterbi)
    """
    assert not return_best_sequence or (viterbi and not return_alpha)

    def recurrence(obs, previous, transitions):
        previous = previous.dimshuffle(0, 'x')
        obs = obs.dimshuffle('x', 0)
        if viterbi:
            scores = previous + obs + transitions
            out = scores.max(axis=0)
            if return_best_sequence:
                out2 = scores.argmax(axis=0)
                return out, out2
            else:
                return out
        else:
            return log_sum_exp(previous + obs + transitions, axis=0)

    initial = observations[0]
    alpha, _ = theano.scan(
        fn=recurrence,
        outputs_info=(initial, None) if return_best_sequence else initial,
        sequences=[observations[1:]],
        non_sequences=transitions
    )

    if return_alpha:
        return alpha
    elif return_best_sequence:
        sequence, _ = theano.scan(
            fn=lambda beta_i, previous: beta_i[previous],
            outputs_info=T.cast(T.argmax(alpha[0][-1]), 'int32'),
            sequences=T.cast(alpha[1][::-1], 'int32')
        )
        sequence = T.concatenate([sequence[::-1], [T.argmax(alpha[0][-1])]])
        return sequence
    else:
        if viterbi:
            return alpha[-1].max(axis=0)
        else:
            return log_sum_exp(alpha[-1], axis=0)
